{*******************************************************}
{                                                       }
{       System Popup menu                               }
{       Author Serge Egorov                             }
{                                                       }
{       Work like standard popup menus                  }
{                                                       }
{*******************************************************}

unit SysMenu;

interface

uses Classes, SysUtils, Windows, Menus, Messages, Forms, Dialogs, Controls;

type
  TSysPopupMenu = class(TPopUpMenu)
  private
    FOriginElements: Integer;
    FOldAppOnMessage: TMessageEvent;
    procedure MergeWithSystemMenu;
    procedure AppOnMessage (var Msg: TMsg; var Handled: Boolean);
    procedure Loaded; override;
  public
    destructor Destroy; override;
    function HandleSystemMenu (var Msg: TMsg): boolean;
  end;

procedure Register;

implementation

procedure TSysPopupMenu.MergeWithSystemMenu;
var
  hSysMenu: THandle;
  I: Integer;

  procedure AppendTo(Item: TMenuItem; Menu: HMENU);
  const
    IBreaks: array[TMenuBreak] of DWORD = (MFT_STRING, MFT_MENUBREAK, MFT_MENUBARBREAK);
    IChecks: array[Boolean] of DWORD = (MFS_UNCHECKED, MFS_CHECKED);
    IDefaults: array[Boolean] of DWORD = (0, MFS_DEFAULT);
    IEnables: array[Boolean] of DWORD = (MFS_DISABLED or MFS_GRAYED, MFS_ENABLED);
    IRadios: array[Boolean] of DWORD = (MFT_STRING, MFT_RADIOCHECK);
    ISeparators: array[Boolean] of DWORD = (MFT_STRING, MFT_SEPARATOR);
    IOwnerDraw: array[Boolean] of DWORD = (MFT_STRING, MFT_OWNERDRAW);
  var
    MenuItemInfo: TMenuItemInfo;
    MenuCaption: string;
    NewFlags: Integer;
    IsOwnerDraw: Boolean;
    ParentMenu: TMenu;
  begin
    if Item.Visible then
      with Item do
      begin
        MenuCaption := Caption;
        if Count > 0 then
          MenuItemInfo.hSubMenu := Handle
        else
        if (ShortCut <> scNone) and
           ((Parent = nil) or (Parent.Parent <> nil) or
           (not (Parent.Owner is TMainMenu))) then
          MenuCaption := MenuCaption + #9 + ShortCutToText(ShortCut);
        if Lo(GetVersion) >= 4 then
        begin
          MenuItemInfo.cbSize := 44; // Required for Windows 95
          MenuItemInfo.fMask := MIIM_CHECKMARKS or
                                MIIM_DATA or
                                MIIM_ID or
                                MIIM_STATE or
                                MIIM_SUBMENU or
                                MIIM_TYPE;
          ParentMenu := GetParentMenu;
          IsOwnerDraw := Assigned(ParentMenu) and
                         (ParentMenu.OwnerDraw or (GetImageList <> nil)) or
                         Assigned(Bitmap) and
                         (not Bitmap.Empty);
          MenuItemInfo.fType := IRadios[RadioItem] or
                                IBreaks[Break] or
                                ISeparators[Caption = '-'] or
                                IOwnerDraw[IsOwnerDraw];
          MenuItemInfo.fState := IChecks[Checked] or
                                 IEnables[Enabled] or
                                 IDefaults[Default];
          MenuItemInfo.wID := Command;
          MenuItemInfo.hSubMenu := 0;
          MenuItemInfo.hbmpChecked := 0;
          MenuItemInfo.hbmpUnchecked := 0;
          MenuItemInfo.dwTypeData := PChar(MenuCaption);
          if Count > 0 then
            MenuItemInfo.hSubMenu := Handle;
          InsertMenuItem(Menu, DWORD(-1), True, MenuItemInfo);
        end
        else
        begin
          NewFlags := IBreaks[Break] or
                      IChecks[Checked] or
                      IEnables[Enabled] or
                      ISeparators[Caption = '-'] or
                      MF_BYPOSITION;
          if Count > 0 then
            InsertMenu(Menu, DWORD(-1), MF_POPUP or NewFlags, GetHandle, PChar(MenuCaption))
          else
            InsertMenu(Menu, DWORD(-1), NewFlags, Command, PChar(MenuCaption));
        end;
      end;
  end;

begin
// For prevent changing in Design mode
  if (csDesigning in ComponentState) then
    exit;

  GetSystemMenu((Owner as TForm).Handle, False);
  hSysMenu := GetSystemMenu(TForm(Owner).Handle, False);


// For changing in Design mode
{
  if FOriginElements = 0 then
    FOriginElements := GetMenuItemCount(hSysMenu);

  if FOriginElements <> 0 then
    while GetMenuItemCount(hSysMenu) > FOriginElements do
      DeleteMenu(hSysMenu, FOriginElements+1, MF_BYPOSITION);
}
  for i := 0 to Items.Count-1 do
    AppendTo(Items[i], hSysMenu);
end;

function TSysPopupMenu.HandleSystemMenu;

  function HandleMenuItem(Item: TMenuItem): boolean;
  var
    I: Integer;
  begin
    Result := False;
    for I := 0 to Item.Count-1 do
    begin
      if Item.Items[I].Command = LoWord(Msg.wParam) then
      begin
        Item.Items[I].Click;
        Result := True;
        break;
      end
      else
        HandleMenuItem(Item.Items[I]);
    end;
  end;

begin
  Result := False;
  inherited;
  HandleMenuItem(Items);
end;

procedure TSysPopupMenu.AppOnMessage (var Msg: TMsg; var Handled: Boolean);
begin
  if Assigned(FOldAppOnMessage) then
    FOldAppOnMessage(Msg, Handled);
  if (not Handled) and
     (Msg.message = WM_SYSCOMMAND) then
    Handled := HandleSystemMenu (Msg);
end;

destructor TSysPopupMenu.Destroy;
begin
  Application.OnMessage := FOldAppOnMessage;
  inherited Destroy;
end;

procedure TSysPopupMenu.Loaded;
begin
  FOldAppOnMessage := Application.OnMessage;
  Application.OnMessage := AppOnMessage;
  inherited Loaded;
  MergeWithSystemMenu;
end;

procedure Register;
begin
  RegisterComponents ('Standard', [TSysPopupMenu]);
end;

end.
